
(*
We now proceed to the proof of the instance of Parry that we used for our problem.
This module is mostly written in
purely functional style, without side effects or mutable variables. Therefore,
for performance reasons, the encoding of tiles and positions is not the most
naive one: first, tiles are encoded as single integers with
five bit fields of {\tt wgl} bits each
(where variable {\tt wgl} is 5), where the south and west fields are the
two leftmost fields. Functions {\tt withC}, {\tt withN}, {\tt withS}, {\tt withW}, {\tt withE}
return a new integer with the color, north, south, west and east fields changed,
respectively. Functions {\tt color}, {\tt north}, {\tt south}, {\tt west}, {\tt east}
return the color, south, west and east field.

Positions are encoded as two 16 bits fields, where the left field is the x component,
and the right one is the y component.

This allows to encode assemblies and tilesets in map data structures, which
are purely functional (i.e. non-mutable).

*)
module Pats=Parry_client.Client(Job)
open Parry_common
module IntSet=Set.Make (struct type t=int let compare=compare end)

let key=
  let open Cryptokit.RSA in
  {Parry_common.blank_key with size=(String.length Key.n) lsl 3; n=Key.n;d=Key.d;e=Key.e }

open Pats
open Job

(*

\begin{lemma}\label{lem:directed}
Given a tileset $\mathtt{tiles}$, $\mathtt{isDirected}\ \mathtt{tiles}=\mathtt{true}$ if and only if
$\mathtt{tiles}$ is a directed DTAS tileset, that is, no two tiles in $\mathtt{tiles}$ have the same input (i.e. south and west) glues.

\begin{proof}

$\mathtt{isDirected}$ works by traversing the tileset using {\tt IntMap.fold},
with a set accumulator. Let $t_1,\ldots t_n$ the successive tiles it sees, and
$s_1,\ldots,s_n$ the corresponding accumulator values, and $s_{n+1}$ the final
accumulator value.

We first prove by induction on $i$ that
$s_{i+1}=\{(\mathrm{South}(t_j),\mathrm{West}(t_j) | j\leq i\}$
if and only if
no two tiles, among $t_1,t_2,\ldots,t_i$, have the same south and west glues.


Since $s_1=\emptyset$, this holds for $i=1$.
For $i\geq 2$, by the induction hypothesis,
if $(\mathrm{South}(t_i),\mathrm{West}(t_i)) \not\in s_i$,
then $t_i$ has different south and west glues from all $t_j$ for $j<i$, in which
case $s_{i+1}=s_i\cup\{(\mathrm{South}(t_i),\mathrm{West}(t_i))\}$.

Else, two tiles have the same south and west glues; thus, exception {\tt Not\_directed}
is raised, and {\tt isDirected ts} is {\tt false}.

Therefore, the induction hypothesis holds also for $i+1$.

By induction, we conclude that if no exception has been raised,
$s_n$ is defined, and thus
the south and west glues of all tiles in {\tt ts} are different, and
{\tt isDirected ts} is {\tt true}.
*)

exception Not_directed
let isDirected ts=
  try
    let _=IntMap.fold (fun _ t s->
      let k=t lsr (3*wgl) in
      if IntSet.mem k s then raise Not_directed else IntSet.add k s
    ) ts IntSet.empty
    in
    true
  with
      Not_directed->false

(*
\end{proof}
\end{lemma}
*)


(*
We now proceed to the proof of the {\tt merge} function.
Its formal specification of this function is given by the following Lemma:

\begin{lemma}\label{lem:merge}
Given a tileset $\mathtt{ts}$, a color $\mathtt{c}$, an index into the tileset $\mathtt{i}$,
a pair of north/south glues $\mathtt{a0\_}$ and $\mathtt{b0\_}$,
and a pair of east/west glues $\mathtt{a1\_}$ and $\mathtt{b1\_}$,
the function $\mathtt{merge}$ returns a tileset in which the $\mathtt{i}$th
tile of $\mathtt{ts}$ is set to color $\mathtt{c}$, and all north/south glues
in $\mathtt{ts}$ equal to $\mathtt{max(a0\_,b0\_)}$ are set to
$\mathtt{min(a0\_,b0\_)}$ if this value is strictly smaller than {\tt mgl}
(and left unchanged else),
and all east/west glues in $\mathtt{ts}$ equal to $\mathtt{max(a1\_,b1\_)}$ are
set to $\mathtt{min(a1\_,b1\_)}$ if this value is strictly smaller than {\tt mgl}
(and left unchanged else).
\end{lemma}

\begin{proof}

In $\mathtt{merge}$, the pair $\mathtt{(a0,b0)}$ is formed such that
$\mathtt{a0} = \mathtt{min(a0\_,b0\_)}$ and $\mathtt{b0} =
\mathtt{max(a0\_,b0\_)}$.  Similarly, the pair $\mathtt{(a1,b1)}$ is formed such
that $\mathtt{a1} = \mathtt{min(a1\_,b1\_)}$ and $\mathtt{b1} =
\mathtt{max(a1\_,b1\_)}$.

Then, a new tileset is created, by folding through all the tiles of {\tt ts},
adding the modified tiles to an accumulator tileset {\tt ts'}. The modification
clearly follows our claim.
*)
let merge a0_ b0_ a1_ b1_ i0 c ts=
  let (a0,b0)=if a0_<b0_ then (a0_,b0_) else (b0_,a0_) in
  let (a1,b1)=if a1_<b1_ then (a1_,b1_) else (b1_,a1_) in
  IntMap.fold (fun i t ts'->
    let u=if i=i0 then withC t c else t in
    let v=if south u=b0 && a0<mgl then withS u a0 else u in
    let w=if west v=b1 && a1<mgl then withW v a1 else v in
    let x=if north w=b0 && a0<mgl then withN w a0 else w in
    let y=if east x=b1 && a1<mgl then withE x a1 else x in
    if y=t then ts' else IntMap.add i y ts'
  ) ts ts

(*
\end{proof}

The next step of our proof is to prove the core computation, called {\tt placeTile}.
We first need to introduce the pattern, hardcoded in the program, 
*)

let pattern=
  [|[|1;0;1;1;0;1;1;1;0;1;1;1;1;1;1;0;1;1;1;1;1|];
    [|0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;1;1;0;1;1;0;1;1;1;1;1;1;0;1;1;1;1|];
    [|0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;1;1;1;1;1;0;1;1;1;1;1;1;0;1;1;1;1|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;1;1;1;1;1;0;1;1;1;1;1;1;0;1;1;1;1|];
    [|0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;1;1;1;1;1;0;1;1;1;1;1;1;0;1;1;1;1|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;1;1;1;1;1;0;1;1;1;1;1;1;0;1;1;1;1|];
    [|0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;0;1;0;1;1;1;1;1;1;1;0;1;1;1;1;1;0;1;1;1|];
    [|0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;0;1;1;1;1;1;1;1;0;1;1;1;1;1;1;0;1|];
    [|0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0;0|];
    [|0;1;1;1;1;0;1;0;1;1;0;1;1;0;1;1;1;1;0;1;1|];
    [|1;1;0;1;0;1;1;1;1;1;1;1;0;1;1;1;1;1;1;1;1|]|]

(*


Finally, the core of our algorithm is the following recursive function, $\mathtt{placeTile}$, which moves through all locations in the pattern in the ordering shown by Figure~\ref{fig:order} and places tiles from the current tile set (often modifying the tile set, too) as long as it is able to.  By making recursive calls which attempt all possibilities, it ensures that the full set of possible tile sets (up to isomorphism) is explored and returns exactly those which self-assemble the given pattern. The arguments to $\mathtt{placeTile}$ are:

\begin{enumerate}
    \item[] {\tt share}: a boolean value set by the server, possibly telling the client to ``share'' the current job
    \item[] {\tt save}: a function that we have explained in Lemma \ref{lem:dowork}, that {\tt placeTile} can use to ``save'' intermediate results in case it is asked to reshare, or killed (for instance if it runs on a cluster).
    \item[] {\tt results}: a list of results that have been found so far.
    \item[] {\tt jobs}: a list of jobs to treat. Each job contains four relevant fields for the actual computation:
\begin{enumerate}
    \item[] {\tt posX},{\tt posY}: the coordinates of the current position in the assembly which $\mathtt{placeTile}$ should attempt to fill with a tile

    \item[] {\tt tileset}: the current tileset (which is a vector of integer values which are each the concatenated integer values representing the properties of a tile type)

    \item[] {\tt assembly}: the current assembly (which is a two-dimensional vector storing the index of the tile type, in the tileset, which is located at each pair of $(x,y)$ coordinates
\end{enumerate}
\end{enumerate}

Note that $\mathtt{placeTile}$ also makes use of the globally defined
two-dimensional vector $\mathtt{pattern}$ which, at each location representing a
pair of $(x,y)$ coordinates, defines one of two colors (i.e. 0 (black) or 1
(white)) for the pattern at that location.

\begin{lemma}\label{lem:placeTile}

For any list of jobs $\mathtt{j}_{\mathtt 0}$, any value of {\tt share} and any function {\tt
save}, {\tt placeTile share save [] $\mathtt{j}_{\mathtt 0}$} returns the list of all subjobs of jobs of
$\mathtt{j_0}$ that have not been explored, and all results that have been found during
the exploration of the explored subjobs of $\mathtt{j}_{\mathtt 0}$.

Moreover, all its calls to {\tt save} are of the form {\tt save j r}, where
{\tt j} is the list of all subjobs of $\mathtt{j}_{\mathtt 0}$ that have not been completely
explored, and $\mathtt{r}$ is the list of all results that have been found in the
exploration of all other subjobs of $\mathtt{j}_{\mathtt 0}$.

\end{lemma}

%Beginning from assembly $\mathtt{assemb}$, tileset $\mathtt{tiles}$, and location $(x,y)$, if $\mathtt{share} = \mathtt{false}$ the function $\mathtt{placeTiles}$ returns a list of all possible pairs of completed assemblies and the tilesets of size $\leq 13$ which can self-assemble them.


%If $\mathtt{share} = \mathtt{true}$, $\mathtt{placeTile}$ creates a list of ``jobs'' which continue from the next tile location and sends them to the server, which will ensure that they will be completed by some client.  The server and the mechanism for distributing jobs is described in Sections \ref{subsect:server} and \ref{subsect:client}.

\begin{proof}

We will prove, by induction on the number of subjobs of
$\mathtt{j}_{\mathtt{0}}$, that for all values of $\mathtt{r}$ and {\tt j},
{\tt placeTile share save r j} is the concatenation of {\tt r} with all the results
found in the exploration, and all the subjobs of jobs of {\tt j} that have not
been explored.

Moreover, we will prove the following invariant:

\hfill

\begin{invariant}
\label{l}

The recursive calls of {\tt placeTile} are all such that the {\tt results} and
{\tt jobs} arguments verify the condition that {\tt jobs} is the list of all
subjobs of the initial job list that have not been explored and contains no
results, and {\tt results} is the list of all results that have been found
during the exploration of all other subjobs of the initial job list.

\end{invariant}

*)

let pos a b=(a lsl 16) lor b


let rec placeTile_ posX posY tileset assembly=
  let inS=if posY<=0 then mgl else
      try
        north (
          IntMap.find
            (IntMap.find (pos posX (posY-1)) assembly)
            tileset
        )
      with
          Not_found-> mgl
  in
  let inW=if posX<=0 then mgl else
      try
        east (
          IntMap.find
            (IntMap.find (pos (posX-1) posY) assembly)
            tileset
        )
      with
          Not_found-> mgl
  in
  if inS=mgl && posY>0 then
    if posX < Array.length pattern.(0) then
      placeTile_ posX 0 tileset assembly
    else
      if posY+1 >= Array.length pattern then
        [{posX=posX;posY=posY;tileset=tileset;assembly=assembly}]
      else
        placeTile_ 0 (posY+1) tileset assembly
  else
    if inW=mgl && posX>0 then
      if posY < Array.length pattern then
        placeTile_ 0 posY tileset assembly
      else
        if posX+1 >= Array.length pattern.(0) then
          [{posX=posX;posY=posY;tileset=tileset;assembly=assembly}]
        else
          placeTile_ (posX+1) 0 tileset assembly
    else
      let nextX,nextY=if posY>posX then (posX+1,posY) else (posX,posY+1)
      and col=pattern.(posY).(posX)
      and key=((inS lsl wgl) lor inW) in
      let possible=
        IntMap.fold (fun k a x->
          if a lsr (3*wgl)=key then k else x
        ) tileset (-1)
      in
      if possible>=0 then
        let tile=IntMap.find possible tileset in
        if color tile = col || color tile=mgl then
          placeTile_ nextX nextY
            (IntMap.add possible (withC tile col) tileset)
            (IntMap.add (pos posX posY) possible assembly)
        else
          []
      else

(*

Or there is no matching tile, in which case we simply try all tiles that can be placed
at the current position, which fall in either of two cases:

\begin{itemize}

\item tiles whose color matches the pattern's color at the current position.
\item tiles that have not yet been used, i.e. whose color is not yet defined.
We only need to consider one of them, because we do not consider solutions that are
equivalent by renaming. Therefore, we use the {\tt seenBlank} argument of {\tt
tryAllTiles} to keep track of whether we have already tried an uncolored tile.

\end{itemize}

In both cases, we merge these tiles' west and south glues with {\tt inW} and
{\tt inS}, respectively: according to Lemma \ref{lem:merge}, this means that we
adjust the tileset so that the chosen tile can be placed without mismatches at
the current position.

Formally, we can easily prove by induction on {\tt i}, that {\tt tryAllTiles i
False l} (respectively {\tt tryAllTiles i True l}) is the list of all subjobs of
{\tt j}, that try to place a tile of index {\tt i} or larger in the tileset, not
including (respectively including) uncolored tiles, along with the jobs of list
{\tt l}.
*)
        snd (
          IntMap.fold (fun k t (tried_blank,results)->
            if color t=col || (color t=mgl && not tried_blank) then (
              let merged=merge inS (south t) inW (west t) k col tileset in
              if isDirected merged then
                (tried_blank || color t=mgl,
                 (placeTile_ nextX nextY merged (IntMap.add (pos posX posY) k assembly))
                 @results)
              else (
                (tried_blank || color t=mgl,results)
              )
            ) else
              (tried_blank,results)
          ) tileset (false,[])
        )



let counter=ref 0
let rec placeTile share save results js=
  match js with
      []->

(*
The first case is when the list of jobs to explore is empty. In this case,
we simply return the list of found results, and the claim holds.
*)

        (results,[])

    | j::s->(

(*
The following expression is the only place where {\tt placeTile} calls {\tt save}.
Although it is not purely functional, the arguments {\tt results}
and {\tt js} to {\tt save} are non-mutable, and hence not modified by this call.
Moreover, by invariant \ref{l} on the recursive calls of {\tt placeTile}, our claim
on the calls to {\tt save} clearly holds.
*)

      if !counter>=10000 then (
        save results js;
        counter:=0;
      ) else incr counter;

(*
First, the variable $\mathtt{inS}$ is the glue to the south of the location
$(x,y)$ to be tiled (i.e. its south input).  If the location to the south is
outside of the pattern or if there is no tile there, then the glue value of $\mathtt{mgl}$
is used.  In an analogous manner, the variable $\mathtt{inW}$ is the value of
the input glue to the west.
*)

      let inS=if j.posY<=0 then mgl else
          try
            north (
              IntMap.find
                (IntMap.find (pos j.posX (j.posY-1)) j.assembly)
                j.tileset
            )
          with
              Not_found-> mgl
      in
      let inW=if j.posX<=0 then mgl else
          try
            east (
              IntMap.find
                (IntMap.find (pos (j.posX-1) j.posY) j.assembly)
                j.tileset
            )
          with
              Not_found-> mgl
      in

(*

If {\tt inS} (respectively {\tt inW}) is {\tt mgl}, and we are not on the south
(respectively west) border, then there is no more tile we can add on the current
row (respectively column). Therefore, we must start a new column (respectively a
new row). Remark that since we keep alternating between adding rows and columns,
we maintain the following invariant: $\mathtt{posX}\geq\mathtt{posY}$ if and only if we
are adding a new row.
This is what the following code does. Invariant \ref{l}, on {\tt placeTile}'s recursive
calls, is clearly preserved by all the calls in this portion of the code.
*)
      if inS=mgl && j.posY>0 then
        if j.posX < Array.length pattern.(0) then
          placeTile share save results ({ j with posY=0 }::s)
        else
          if j.posY+1 >= Array.length pattern then
            placeTile share save (j::results) s
          else
            placeTile share save results ({ j with posX=0; posY=j.posY+1 }::s)
      else
        if inW=mgl && j.posX>0 then
          if j.posY < Array.length pattern then
            placeTile share save results ({ j with posX=0 }::s)
          else
            if j.posX+1 >= Array.length pattern.(0) then
              placeTile share save (j::results) s
            else
              placeTile share save results ({ j with posY=0; posX=j.posX+1 }::s)
        else

(*
Else, both the south and west glues are defined, or we are at the beginning of a
row or a column. Hence, there are two possible cases: either there is already a
tile with matching south and west glues, or there is none.  In the first case,
we have no choice but to place that tile at the current position, and move on to
the next position, which is done when $\mathtt{possible}\geq 0$:
*)

          let nextX,nextY=if j.posY>j.posX then (j.posX+1,j.posY) else (j.posX,j.posY+1)
          and col=pattern.(j.posY).(j.posX)
          and key=((inS lsl wgl) lor inW) in
          let possible=
            IntMap.fold (fun k a x->
              if a lsr (3*wgl)=key then k else x
            ) j.tileset (-1)
          in
          if possible>=0 then
            let tile=IntMap.find possible j.tileset in
            if color tile = col || color tile=mgl then
              placeTile share save results
                ({ posX=nextX;posY=nextY;
                   tileset=IntMap.add possible (withC tile col) j.tileset;
                   assembly=IntMap.add (pos j.posX j.posY) possible j.assembly }
                 :: s)
            else (
              placeTile share save results s
            )
          else

(*

Or there is no matching tile, in which case we simply try all tiles that can be placed
at the current position, which fall in either of two cases:

\begin{itemize}

\item tiles whose color matches the pattern's color at the current position.
\item tiles that have not yet been used, i.e. whose color is not yet defined.
We only need to consider one of them, because we do not consider solutions that are
equivalent by renaming. Therefore, we use the {\tt seenBlank} argument of {\tt
tryAllTiles} to keep track of whether we have already tried an uncolored tile.

\end{itemize}

In both cases, we merge these tiles' west and south glues with {\tt inW} and
{\tt inS}, respectively: according to Lemma \ref{lem:merge}, this means that we
adjust the tileset so that the chosen tile can be placed without mismatches at
the current position.

Formally, we can easily prove by induction on {\tt i}, that {\tt tryAllTiles i
False l} (respectively {\tt tryAllTiles i True l}) is the list of all subjobs of
{\tt j}, that try to place a tile of index {\tt i} or larger in the tileset, not
including (respectively including) uncolored tiles, along with the jobs of list
{\tt l}.
*)
            let _,next_jobs=
              IntMap.fold (fun k t (tried_blank,jobs)->
                if color t=col || (color t=mgl && not tried_blank) then (
                  let merged=merge inS (south t) inW (west t) k col j.tileset in
                  if isDirected merged then
                    (tried_blank || color t=mgl,
                     { posX=nextX;posY=nextY;
                       tileset=merged;
                       assembly=IntMap.add (pos j.posX j.posY) k j.assembly }::jobs)
                  else (
                    (tried_blank || color t=mgl,jobs)
                  )
                ) else
                  (tried_blank,jobs)
              ) j.tileset (false,s)
            in

(*
From our proof of the {\tt tryAllTiles} function, the following recursive
call to {\tt placeTile} preserves the invariant on {\tt placeTile}'s recursive
calls. Indeed, {\tt results} has not changed, and {\tt nextJobs} now contains
{\tt js}, along with all subjobs of {\tt j} (up to renaming of unused tiles).
By the definitions of jobs (see Definition
\ref{def:jobs}), invariant \ref{l} is clearly preserved.
*)

            placeTile share save results next_jobs
    )

(*
\end{proof}
*)

(*
The last part of this module uses the client framework proven in Section
\ref{subsect:client} to call the {\tt placeTile} function.
*)

let _=
  Pats.client
    { server=(Unix.gethostbyname "pats.lif.univ-mrs.fr").Unix.h_addr_list.(0);
      port=5129;
      key=key }
    (fun save j->
      placeTile true save [] [j]
    )
